# Tool-Naranayaswamy-Moynihan-Hodge model
# by Florian Pabst 2020


import re
import os
import sys
import time

import numpy as np
import matplotlib.pyplot as plt

from scipy.integrate import quad, cumulative_trapezoid
from scipy.optimize import curve_fit, fsolve
from scipy.stats import linregress

from nmreval.data import Points


def slope(x, t, m):
    return t+x*m


def tau_k(tau_0, deltaE, temp_k, xx, T_f_km1):
    return tau_0 * np.exp(xx*deltaE / (R*temp_k) + (1-xx) * deltaE / (R*T_f_km1))


# Read data
dataName = sys.argv[1]
try:
    data = np.loadtxt(dataName, skiprows=0).T
    # print("Loading")
    temp = data[0]
    heat_capacity = data[1]
except IndexError:
    # print("File not found")
    exit()


def get_fictive_temperature(pts: Points, glass: tuple[float, float], liquid: tuple[float, float]):
    min_glass, max_glass = min(glass), max(glass)
    min_liquid, max_liquid = min(liquid), max(liquid)

    region = pts.copy()
    region.cut(min_glass, max_liquid)

    glass_regime = (min_glass < region.x) & (region.x < max_glass)
    regress = linregress(region.x[glass_regime], region.y[glass_regime])
    glass_extrapolation = regress.slope * region.x + regress.intercept

    region.y -= glass_extrapolation

    liquid_regime = (min_liquid < region.x) & (region.x < max_liquid)
    regress = linregress(region.x[liquid_regime], region.y[liquid_regime])
    liquid_extrapolation = regress.slope * region.x + regress.intercept

    real_area = -cumulative_trapezoid(region.y, region.x, initial=0)
    real_area -= real_area[-1]
    equivalent_area = cumulative_trapezoid(liquid_extrapolation, region.x, initial=0)
    equivalent_area -= equivalent_area[-1]
    equivalent_area *= -1

    return region.x, region.x[np.argmin(np.abs(real_area[:, None] - equivalent_area[None, :]), axis=1)]


curve = Points(x=temp, y=heat_capacity, name=dataName)

# First step: Calculate fictive Temperature T_f(T) from Cp data, then dT_f/dT

temperatures, fictiveTemp = get_fictive_temperature(curve, glass=(150, 160), liquid=(190, 200))

# Determine limiting T_f, i.e. T_f at lowest temperature
initial_Tf = np.mean(fictiveTemp[:20])

rate = curve.value

# Calculate dTf/dT
with np.errstate(all='ignore'):
    dTfdT = np.gradient(fictiveTemp, temperatures)
nan_filter = ~np.isnan(dTfdT)
temperatures = np.array(temperatures)[nan_filter]
dTfdT = dTfdT[nan_filter]


exponents = []
taus = []
temps = [0]
T_f_ns = []
R = 8.314


def tnmfunc(temperature, tau_g, x, beta, deltaE):
    step = len(temperature)

    Tf = np.empty(step)
    Tf[0] = temperature[0]

    delT = np.diff(temperature)
    delt = np.abs(delT) * 60 / rate

    tau = np.empty(step)
    dttau = np.zeros(step)
    temp_0 = temperature[0]

    for i in range(0, step-1):
        tau[i] = relax(temperature[i+1], Tf[i], tau_g, T_g, deltaE, x)
        dttau[:i] += delt[i] / tau[i]
        Tf[i+1] = np.sum(delT[:i] * (1-np.exp(-dttau[:i]**beta))) + temp_0

    return Tf


def relax(t, tf, tau_g, t_glass, ea, x):
    h = ea/R
    return tau_g * np.exp((x*h / t) + ((1 - x) * h / tf) - h / t_glass)


# TNMH Code (adapted from Badrinarayanan's PhD Thesis):
def tnmfunc2(xdata, tau_0, xx, beta, deltaE):
    delhR = deltaE/8.314
    T = []
    Tf = []
    t = []
    delt = []
    tau = []
    dttau = []
    delT = []
    step = len(xdata)
    for k in range(0, step):
        dttau.append(0)
        T.append(240)
        Tf.append(240)
        t.append(0)
        delt.append(0)
        tau.append(0)
        dttau.append(0)
        delT.append(0)
    T[1] = xdata[1]
    Tf[1] = T[1]
    t[1] = 0
    delt[1] = 0
    tau[1] = tau_0 * np.exp((xx*delhR / T[1]) + ((1 - xx)*delhR / Tf[1])-delhR/T_g)
    dttau[1] = delt[1] / tau[1]
    delT[1] = 0

    Tfinit = T[1]
    T_f_ns = [xdata[0]]
    for i in range(1, step):
        T[i] = xdata[i]
        delT[i] = xdata[i] - xdata[i-1]
        delt[i] = abs(delT[i]) / (rate/60)
        t[i] = t[i-1] + delt[i]
        tau[i] = tau_0 * np.exp((delhR*xx / T[i]) + ((1 - xx)*delhR / Tf[i-1])-delhR/T_g)
        # print(tau[i], T[i], Tf[i-1], dttau)
        for j in np.arange(2, i).reshape(-1):
            dttau[j] = dttau[j] + (delt[i] / tau[i])
            Tfinit = Tfinit + (delT[j] * (1 - np.exp(- (dttau[j] ** beta))))
            Tf[i] = Tfinit
        T_f_ns.append(Tfinit)
        Tfinit = Tf[1]

    return T_f_ns


def tnmfunc3(xdata, tau_0, xx, beta, deltaE):
    delhR = deltaE/8.314

    step = len(xdata)

    dttau = []
    T = []
    Tf = []
    t = []
    delT = []
    delt = []
    tau = []
    for _ in range(step):
        T.append(np.nan)
        Tf.append(np.nan)
        t.append(np.nan)
        delT.append(np.nan)
        delt.append(np.nan)
        tau.append(np.nan)
        dttau.append(0)

    T[0] = xdata[0]
    Tf[0] = T[0]
    t[0] = 0
    delT[0] = 0
    delt[0] = 0
    tau[0] = tau_0 * np.exp(xx * delhR/T[0] + (1-xx)*delhR/Tf[0] - delhR/T_g)
    dttau[0] = delt[0]/tau[0]

    Tfinit = T[0]

    for i in range(1, step):
        T[i] = xdata[i]
        delT[i] = xdata[i] - xdata[i-1]
        delt[i] = abs(delT[i]) * 60 / rate
        t[i] = t[i-1] + delt[i]
        tau[i] = tau_0 * np.exp(xx * delhR/T[i] + (1-xx)*delhR/Tf[i-1] - delhR/T_g)

        for j in range(1, i):
            dttau[j] += delt[i]/tau[i]
            Tfinit += delT[j] * (1-np.exp(-dttau[j]**beta))

        Tf[i] = Tfinit
        Tfinit = T[0]

    return Tf


def tnmfunc2(xdata, tau_0, xx, beta, deltaE):
    delhR = deltaE/8.314
    T = []
    Tf = []
    t = []
    delt = []
    tau = []
    dttau = []
    delT = []
    step = len(xdata)
    for k in range(0, step):
        dttau.append(0)
        T.append(240)
        Tf.append(240)
        t.append(0)
        delt.append(0)
        tau.append(0)
        dttau.append(0)
        delT.append(0)
    T[1] = xdata[1]
    Tf[1] = T[1]
    t[1] = 0
    delt[1] = 0
    tau[1] = tau_0 * np.exp((xx*delhR / T[1]) + ((1 - xx)*delhR / Tf[1])-delhR/T_g)
    dttau[1] = delt[1] / tau[1]
    delT[1] = 0

    Tfinit = T[1]
    T_f_ns = [xdata[0]]
    for i in range(1, step):
        T[i] = xdata[i]
        delT[i] = xdata[i] - xdata[i-1]
        delt[i] = abs(delT[i]) / (rate/60)
        t[i] = t[i-1] + delt[i]
        tau[i] = tau_0 * np.exp((delhR*xx / T[i]) + ((1 - xx)*delhR / Tf[i-1])-delhR/T_g)
        # print(tau[i], T[i], Tf[i-1], dttau)
        for j in np.arange(2, i).reshape(-1):
            dttau[j] = dttau[j] + (delt[i] / tau[i])
            Tfinit = Tfinit + (delT[j] * (1 - np.exp(- (dttau[j] ** beta))))
            Tf[i] = Tfinit
        T_f_ns.append(Tfinit)
        Tfinit = Tf[1]

    return T_f_ns


# Use T_g = T_fictive for tau determination
T_g = initial_Tf

# Start values
p0 = [10, 0.5, 0.4, 225085]


def fitTNMH(x, tau_0, xx, beta, deltaE):
    dTNMHdT = []
    modelTemp = np.append(x[::-1], x[1:])

    TNMH = tnmfunc2(modelTemp, tau_0, xx, beta, deltaE)
    for i in range(len(modelTemp)-1):
        dTNMHdT.append((TNMH[i+1]-TNMH[i]) / (modelTemp[i+1]-modelTemp[i]))
    res = dTNMHdT[len(modelTemp)//2-1:]

    plt.plot(x, res)
    plt.show()

    return np.array(res)  


# Use only every 20th data point to reduce fitting time
temperaturesInterpol = np.linspace(temperatures[0], temperatures[-1], 20)
temperaturesInterpol = np.linspace(150, 200, 51)

temperaturesInterpol = np.append(temperaturesInterpol[::-1], temperaturesInterpol[1:])

start = time.time()
tnmfunc2(temperaturesInterpol, *p0)
print('Flo', time.time()-start)
start = time.time()
tnmfunc3(temperaturesInterpol, *p0)
print('list', time.time()-start)
start = time.time()
tnmfunc(temperaturesInterpol, *p0)
print('Array', time.time()-start)

dTfdTInterpol = np.interp(temperaturesInterpol, temperatures, dTfdT)

plt.plot(tnmfunc2(temperaturesInterpol, *p0), label='Flo')
plt.plot(tnmfunc3(temperaturesInterpol, *p0), label='List')
plt.plot(tnmfunc(temperaturesInterpol, *p0), label='array')
plt.legend()
plt.show()

plt.plot(np.diff(tnmfunc2(temperaturesInterpol, *p0)), label='Flo')
plt.plot(np.diff(tnmfunc3(temperaturesInterpol, *p0)), label='List')
plt.plot(np.diff(tnmfunc(temperaturesInterpol, *p0)), label='array')
plt.legend()
plt.show()
# print(fitTNMH(temperaturesInterpol, *p0))
